import netgen.gui
import matplotlib.pyplot as plt
from make_1d_mesh import *
import gc

import numpy as np
import matplotlib as mpl
from ngsolve import *
# mesh and spaces initialization

def mesh_and_spaces(nel,length,ord):
    mesh = Mesh(Make1DMesh([[0,length]],[nel],bc=[1,2]))
    #m2 , n = SegMesh_refined(nel,0.0,length,1200)
    #m2 = SegMesh(nel,0,3600,0)
    #m2.dim = 1
   
    #mesh = Mesh(m2)
    Y = L2(mesh, order=ord)  # space can be fixed for the whole routine
    Z = FacetFESpace(mesh, order=0,dirichlet = [1,2])
    X = FESpace([Y,Z,Y,Z,Y,Z])
    xpoints = np.linspace(0,length,nel+1)
    return X,Y, mesh
# scaling parameters
def scaling():
    lam= 0.0433
    beta= 3.86
    z1= 1
    z2= -1
    kappa = 6
    D1=1.33*pow(10,-9)
    D2= 2.07*pow(10,-9)
    return lam, beta, kappa, z1,z2,D1,D2
def PNP_solver(Vapl,bc1,mesh,length,X,Y,Sig,radius,nel):
    inform  = 0
    # shape function prep
    xpoints = GridFunction(Y)
    xpoints.Set(x)
    R = GridFunction(Y)
    R.Set(radius)
    #  surface charge function prep
    Sigma= GridFunction(Y)
    Sigma.Set(Sig)
    lamb, beta,kappa,z1,z2,D1,D2 = scaling()
    A= GridFunction(Y)
    A = R*R*22/7
    # trial and test functions
    V,V_hat,u1,u1_hat,u2,u2_hat= X.TrialFunction()
    v1,v1_hat,v2,v2_hat,v3,v3_hat= X.TestFunction()
    # trial functions jumps
    jump_V = V-V_hat
    jump_u1 = u1-u1_hat
    jump_u2 = u2-u2_hat
    # test functions jumps
    jump_v1= v1-v1_hat
    jump_v2= v2-v2_hat
    jump_v3= v3-v3_hat
    
    alpha = 4
    order =1
    h = specialcf.mesh_size
    n = specialcf.normal(mesh.dim)
    # bilinear form 
    print('Bilinar form prep')    

    a = BilinearForm(X, symmetric=False)
    # HDG formulation of the PB equation
    a += SymbolicBFI( lamb*A*grad(V) * grad(v1) - A*kappa*z1*exp( u1- z1*beta*V)* v1 \
        - A*kappa*z2* exp(u2- z2*beta*V)*v1  -2*22/7*R*Sigma*v1  ) 
    a += SymbolicBFI(-lamb*A*grad(V)*n*jump_v1, element_boundary=True)
    a += SymbolicBFI(-lamb*A*jump_V*n*v1, element_boundary=True)
    a += SymbolicBFI(alpha*order*order/h *lamb*A* jump_V*jump_v1,element_boundary=True)
    # HDG formulation of the cont equation 
    a += SymbolicBFI(- A*exp(u1 - z1*beta*V)*grad (u1)* grad(v2)) # first cont equation
    a += SymbolicBFI(  A*exp(u1 - z1*beta*V)*grad (u1)*n* jump_v2, element_boundary=True) # first cont equation
    a += SymbolicBFI(  A*exp(u1 - z1*beta*V)*grad (v2)*n*jump_u1, element_boundary=True ) # first cont equation
    a += SymbolicBFI(-alpha*order*order/h * A*exp(u1 - z1*beta*V)*jump_u1*jump_v2,element_boundary=True)
    a += SymbolicBFI(- A*exp(u2 - z2*beta*V)*grad (u2)* grad(v3))# second cont equation
    a += SymbolicBFI(  A*exp(u2 - z2*beta*V)*grad (u2)*n* jump_v3, element_boundary=True) # second cont equation
    a += SymbolicBFI(  A*exp(u2 - z2*beta*V)*grad (v3)*n*jump_u2, element_boundary=True ) # second cont equation
    a += SymbolicBFI(-alpha*order*order/h * A*exp(u2 - z2*beta*V)*jump_u2*jump_v3,element_boundary=True)    
    # boundary conditions  and initial guesses 
    u = GridFunction(X)
    # setting the boundary conditions
    u.components[1].vec[nel]= Vapl
    u.components[3].vec[nel]=z1*beta*u.components[1].vec[nel] + log(bc1)
    u.components[5].vec[nel]=z2*beta*u.components[1].vec[nel] + log(bc1)
    res = u.vec.CreateVector()
    du = u.vec.CreateVector()
    print('bilinear form Done')

    ################################# newton loop
    for it in range(250):      
        #a.Apply(u.vec, res)
        a.Apply(u.vec, res, heapsize=1000000000)
        a.AssembleLinearization(u.vec)    
        inv=a.mat.Inverse(X.FreeDofs(),inverse="umfpack") 
        du.data =inv* res
        u.vec.data -= du
        #stopping criteria
        stopcritval = sqrt(abs(InnerProduct(du,res)))
        print ("Iteration",it)
        print ("<A u",it,", A u",it,">_{-1}^0.5 = ", stopcritval)
        if stopcritval < 1e-9:
            break  
    ################################## current calculations  and output functions    
    c1 = GridFunction(Y)
    c2 = GridFunction(Y)
    Vout = GridFunction(Y)
    Vout.Set(u.components[0])
    c1.Set(exp(u.components[2] - z1*beta*u.components[0]))
    c2.Set(exp(u.components[4] - z2*beta*u.components[0]))

    gradu1= grad(u.components[2])
    gradu2= grad(u.components[4])
    J1 = GridFunction(Y)
    J2 = GridFunction(Y)
    J1.Set(1.33*A*c1*gradu1)
    J2.Set(0.79*A*c2*gradu2)
    e = 1.602*pow(10,-9)
  
    I = pow(10,2)*(-z1*J1.vec[int(nel/2)] - z2*J2.vec[int(nel/2)])
    print('I =', I)
    del a, u, gradu1,gradu2,A,R,Sigma,V,V_hat
    gc.collect()
    return Vout,c1,c2, J1.vec[int(nel/2)], J2.vec[int(nel/2)]



# Make 1D mesh
nel =200
length = 3600
Vap = -2
bc = 1

X,Y,mesh = mesh_and_spaces(nel,length,1)
xpoints = GridFunction(Y)
xpoints.Set(x)

# setting sigma and R
Sigma= GridFunction(Y)
Sigma.Set(1*IfPos(x-1200,1,0)*IfPos(-x+2400,1,0))
rsmall  = 2
rbig = 10
sh =  rsmall + (rbig-rsmall)*(x-1200)/1200 
R = GridFunction(Y)
R.Set((sh)*IfPos(x-1200,1,0)*IfPos(-x+2400,1,0) + 3*rbig*IfPos(x-1200,0,1)+ 3*rbig*IfPos(-x+2400,0,1) )
A = GridFunction(Y)
A = R*R



   
for i in range(10):
    V,c1,c2,J1,J2 = PNP_solver(Vap,bc,mesh,length,X,Y,Sigma,R,nel)

    
    
    
    
